# -*- coding: utf-8 -*-
# Copyright (c) 2015, Vispy Development Team.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.

from .color import UniformColorComponent, VertexColorComponent  # noqa
from .component import VisualComponent  # noqa
from .material import GridContourComponent, ShadingComponent  # noqa
from .normal import VertexNormalComponent  # noqa
from .texture import (TextureComponent, VertexTextureCoordinateComponent,  # noqa
                      TextureCoordinateComponent)  # noqa
from .vertex import XYPosComponent, XYZPosComponent, HeightFieldComponent  # noqa


from .clipper import Clipper  # noqa
from .color2 import Alpha, ColorFilter  # noqa
